#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import json
import time
import unittest

from sst.actions import (
    assert_displayed,
    assert_title,
    click_button,
    fails,
    go_to,
    get_element,
    get_elements,
    get_element_by_css,
    get_elements_by_css,
    get_link_url,
    set_base_url,
    set_window_size,
    wait_for,
    write_textfield,
)
from sst import (
    browsers,
    cases,
)

import deployers


class WebUITest(cases.SSTTestCase):

    browser_factory = browsers.PhantomJSFactory()

    def setUp(self):
        """Sets browser base_url based on the available deployment."""
        super(WebUITest, self).setUp()
        set_base_url('http://{}/'.format(
            deployers.juju_get_ip_and_port('ci-airline-webui-apache')[0]))
        set_window_size(1280, 1024)

    def test_home(self):
        # Explore the open ticket list (home).
        go_to('/')
        assert_title('Ubuntu CI Engine')

        # 'Tickets' is the active link.
        active_link = get_element_by_css('a.active')
        self.assertEqual('Tickets', active_link.text)

        # And the table headers are ...
        dt_header = wait_for(
            get_element_by_css, 'table.yui3-datatable-table thead')
        self.assertEqual(
            ['Title', 'Landing Progress', 'Review Status'],
            [t.text for t in dt_header.find_elements_by_tag_name('th')])

        # A stock YUI paginator is also present.
        paginator = wait_for(get_element_by_css, '#paginator div')
        self.assertEqual(
            'yui3-pagview-container', paginator.get_attribute("class"))

    def test_create_ticket_fails_without_required_fields(self):
        # Creating a ticket without required fields should show errors
        go_to('/create')
        wait_for(get_element, tag='form')
        # Click the submit button without inputting any fields
        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)
        # There should be 3 required fields
        # Either sources or merge proposals is required too, but won't
        # display an error until it gets past this point
        require_errors = get_elements(css_class="error")
        self.assertEqual(3, len(require_errors))
        for error_span in require_errors:
            assert_displayed(error_span)

    def test_create_ticket(self):
        # Create a ticket via WebUI form.
        go_to('/create')
        assert_title('Ubuntu CI Engine')

        wait_for(get_element, tag='form')

        # Submit the form with the required fields.
        title = get_element_by_css('input[name="title"]')
        write_textfield(title, 'Testing WebUI')
        landers = get_element_by_css('input[name="landers"]')
        write_textfield(landers, 'foo@bar.com')
        sources = get_element_by_css('textarea[name="sources"]')
        write_textfield(sources, 'testsource')
        test_notes = get_element_by_css('textarea[name="test_notes"]')
        write_textfield(test_notes, 'test this')
        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)

        # Redirects to the ticket page.
        ticket_title = wait_for(get_element, tag='h2')
        self.assertEqual('Testing WebUI', ticket_title.text)

    def test_create_ticket_with_sources_and_mps(self):
        # Create a ticket using sources and MPs.
        go_to('/create')
        wait_for(get_element, tag='form')

        # Submit the form with the required fields.
        title = get_element_by_css('input[name="title"]')
        write_textfield(title, 'Testing Sources & MPs')
        landers = get_element_by_css('input[name="landers"]')
        write_textfield(landers, 'foo@bar.com')
        sources = get_element_by_css('textarea[name="sources"]')
        write_textfield(sources, 'foo\nbar')
        mps = get_element_by_css('textarea[name="merge_proposals"]')
        write_textfield(
            mps,
            'https://code.launchpad.net/~x/foo/a_branch/+merge/1\n'
            'https://code.launchpad.net/~x/foo/a_branch/+merge/2\n'
            'https://code.launchpad.net/~x/baz/a_branch/+merge/3')
        test_notes = get_element_by_css('textarea[name="test_notes"]')
        write_textfield(test_notes, 'test this')
        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)

        # Redirects to the ticket page where sources & MPs were combined
        # as subtickets. Sources mentioned only in MPs are created.
        ticket_title = wait_for(get_element, tag='h2')
        self.assertEqual('Testing Sources & MPs', ticket_title.text)
        subtickets = get_elements_by_css('table.subticket')
        self.assertEqual(
            [['Source: foo',
              'Status: -',
              'Step: New',
              'Current version: Not available',
              'MPs: https://code.launchpad.net/~x/foo/a_branch/+merge/1',
              'https://code.launchpad.net/~x/foo/a_branch/+merge/2',
              'Artifacts:',
              'Conflicts:'],
             ['Source: bar',
              'Status: -',
              'Step: New',
              'Current version: Not available',
              'MPs:',
              'Artifacts:',
              'Conflicts:'],
             ['Source: baz',
              'Status: -',
              'Step: New',
              'Current version: Not available',
              'MPs: https://code.launchpad.net/~x/baz/a_branch/+merge/3',
              'Artifacts:',
              'Conflicts:']],
            [s.text.splitlines() for s in subtickets])

    def test_create_ticket_fails_without_sources_or_mps(self):
        # Either source packages or merge proposals must be defined
        go_to('/create')
        wait_for(get_element, tag='form')
        # Fill in all the fields except sources and mps
        title = get_element_by_css('input[name="title"]')
        write_textfield(title, 'Testing WebUI')
        landers = get_element_by_css('input[name="landers"]')
        write_textfield(landers, 'foo@bar.com')
        test_notes = get_element_by_css('textarea[name="test_notes"]')
        write_textfield(test_notes, 'test this')
        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)
        # Check that we get the right error
        errors = get_elements(css_class="error")
        self.assertEqual(2, len(errors))
        for error in errors:
            self.assertEqual(
                'Source Packages or Merge Proposals must be supplied',
                error.text)

    def test_edit_ticket(self):
        # Edit a ticket via WebUI form.
        go_to('/')

        # Wait for a rendered ticket list and click on the first
        # available ticket.
        wait_for(get_element_by_css, 'a.yui3-pagview-link-page-active')
        a_ticket_link = get_elements_by_css('a.row_link')[0]
        a_ticket_link.click()

        # Wait for the ticket link to be rendered and click on it.
        edit_link = wait_for(get_element_by_css, 'a.edit-link')
        edit_link.click()

        # Wait for the edit form to be rendered.
        wait_for(get_element, tag='form')

        # bug_id inline validation forbids non-integer inputs and renders
        # a corresponding field error.
        bug = get_element_by_css('input[name="bug_id"]')
        write_textfield(bug, 'aaa')
        title = get_element_by_css('input[name="title"]')
        write_textfield(title, 'just to change focus')
        error = get_element_by_css('span.error')
        self.assertEqual(
            'Please enter a valid Launchpad bug ID.', error.text)

        # An integer bug_id input passes validation and the form can be
        # submitted.
        write_textfield(bug, '1234')
        new_title = 'Edited via the WebUI on {}!'.format(time.time())
        write_textfield(title, new_title)
        fails(get_element_by_css, 'span.error')
        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)

        # Form submission succeeded and redirected back to the ticket
        # page, with the new title and LP bug reference rendered as a link.
        ticket_title = wait_for(get_element, tag='h2')
        self.assertEqual(new_title, ticket_title.text)
        bug_link = get_element_by_css('a.bug-link')
        self.assertEqual('#1234', bug_link.text)
        self.assertEqual(
            'https://launchpad.net/bugs/1234', get_link_url(bug_link))

        # Let's edit the ticket again.
        edit_link = wait_for(get_element_by_css, 'a.edit-link')
        edit_link.click()
        wait_for(get_element, tag='form')

        # Fill the form with an unique title and empty bug value.
        new_title = 'Edited via the WebUI on {}!'.format(time.time())
        title = get_element_by_css('input[name="title"]')
        write_textfield(title, new_title)
        bug = get_element_by_css('input[name="bug_id"]')
        write_textfield(bug, '')
        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)

        # The ticket has a new title and cleared bug reference.
        ticket_title = wait_for(get_element, tag='h2')
        self.assertEqual(new_title, ticket_title.text)
        fails(get_element_by_css, 'a.bug-link')

    def test_edit_ticket_sources_and_mps(self):
        # Edit a ticket sources & MPs.
        go_to('/')

        # Edit the first available ticket.
        wait_for(get_element_by_css, 'a.yui3-pagview-link-page-active')
        a_ticket_link = get_elements_by_css('a.row_link')[0]
        a_ticket_link.click()
        edit_link = wait_for(get_element_by_css, 'a.edit-link')
        edit_link.click()
        wait_for(get_element, tag='form')

        # Update the ticket with a unique title and distinct sources
        # and merge proposals.
        new_title = 'Edited via the WebUI on {}!'.format(time.time())
        title = get_element_by_css('input[name="title"]')
        write_textfield(title, new_title)
        sources = get_element_by_css('textarea[name="sources"]')
        write_textfield(sources, 'boing')
        mps = get_element_by_css('textarea[name="merge_proposals"]')
        write_textfield(
            mps,
            'https://code.launchpad.net/~x/zoing/a_branch/+merge/1\n')
        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)

        # XXX cprov 20141208: existing sources and MPs should be removed
        # upon user confirmation via a modal window before submiting new
        # the edit form.
        ticket_title = wait_for(get_element, tag='h2')
        self.assertEqual(new_title, ticket_title.text)
        subtickets = get_elements_by_css('table.subticket')
        self.assertEqual(
            [['Source: foo',
              'Status: -',
              'Step: New',
              'Current version: Not available',
              'MPs: https://code.launchpad.net/~x/foo/a_branch/+merge/1',
              'https://code.launchpad.net/~x/foo/a_branch/+merge/2',
              'Artifacts:',
              'Conflicts:'],
             ['Source: bar',
              'Status: -',
              'Step: New',
              'Current version: Not available',
              'MPs:',
              'Artifacts:',
              'Conflicts:'],
             ['Source: baz',
              'Status: -',
              'Step: New',
              'Current version: Not available',
              'MPs: https://code.launchpad.net/~x/baz/a_branch/+merge/3',
              'Artifacts:',
              'Conflicts:'],
             ['Source: boing',
              'Status: -',
              'Step: New',
              'Current version: Not available',
              'MPs:',
              'Artifacts:',
              'Conflicts:'],
             ['Source: zoing',
              'Status: -',
              'Step: New',
              'Current version: Not available',
              'MPs: https://code.launchpad.net/~x/zoing/a_branch/+merge/1',
              'Artifacts:',
              'Conflicts:']],
            [s.text.splitlines() for s in subtickets])

    def test_create_first_with_invalid_mps(self):
        # Create a ticket using sources and MPs.
        go_to('/create')
        wait_for(get_element, tag='form')

        # Submit the form with the required fields.
        title = get_element_by_css('input[name="title"]')
        title_text = 'Testing Invalid MPs at {}'.format(time.asctime())
        write_textfield(title, title_text)
        landers = get_element_by_css('input[name="landers"]')
        write_textfield(landers, 'foo@bar.com')
        sources = get_element_by_css('textarea[name="sources"]')
        write_textfield(sources, 'foo\nbar')
        mps = get_element_by_css('textarea[name="merge_proposals"]')
        write_textfield(
            mps,
            'https://code.launchpad.net/~x/foo/a_branch/+merge/1\n'
            'https://code.launchpad.net/~x/foo/a_branch/+merge/abc\n'
            'https://code.launchpad.net/~x/baz/a_branch/+merge/3')
        test_notes = get_element_by_css('textarea[name="test_notes"]')
        write_textfield(test_notes, 'test this')
        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)
        error = get_element_by_css('span.error')
        self.assertEqual(
            'The entry https://code.launchpad.net/~x/foo/a_branch/+merge/abc'
            ' is not a valid MP URL.', error.text)
        write_textfield(
            mps,
            'https://code.launchpad.net/~x/foo/a_branch/+merge/1\n'
            'https://code.launchpad.net/~x/foo/a_branch/+merge/2\n'
            'https://code.launchpad.net/~x/baz/a_branch/+merge/3')
        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)
        # Redirects to the ticket page.
        ticket_title = wait_for(get_element, tag='h2')
        self.assertEqual(title_text, ticket_title.text)

    def test_edit_ticket_with_invalid_mps(self):
        go_to('/create')
        wait_for(get_element, tag='form')

        # Submit the form with the required fields.
        title = get_element_by_css('input[name="title"]')
        title_text = 'Editing with Invalid MPs at {}'.format(time.asctime())
        write_textfield(title, title_text)
        landers = get_element_by_css('input[name="landers"]')
        write_textfield(landers, 'foo@bar.com')
        sources = get_element_by_css('textarea[name="sources"]')
        write_textfield(sources, 'foo\nbar')
        test_notes = get_element_by_css('textarea[name="test_notes"]')
        write_textfield(test_notes, 'test this')

        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)
        # Redirects to the ticket page.
        ticket_title = wait_for(get_element, tag='h2')
        self.assertEqual(title_text, ticket_title.text)

        # Edit a ticket sources & MPs.
        go_to('/')

        # Edit the first available ticket.
        wait_for(get_element_by_css, 'a.yui3-pagview-link-page-active')
        a_ticket_link = get_elements_by_css('a.row_link')[0]
        a_ticket_link.click()
        edit_link = wait_for(get_element_by_css, 'a.edit-link')
        edit_link.click()
        wait_for(get_element, tag='form')

        # Update the ticket with a unique title and distinct sources
        # and merge proposals.
        new_title = 'Edited for testing on {}!'.format(time.asctime())
        title = get_element_by_css('input[name="title"]')
        write_textfield(title, new_title)
        sources = get_element_by_css('textarea[name="sources"]')
        write_textfield(sources, 'boing')
        mps = get_element_by_css('textarea[name="merge_proposals"]')
        write_textfield(
            mps,
            'https://bazaar.launchpad.net/~x/zoing/a_branch/+merge/5\n')
        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)
        error = get_element_by_css('span.error')
        self.assertEqual(
            'The entry https://bazaar.launchpad.net/~x/zoing/a_branch/+merge/5'
            ' is not a valid MP URL.', error.text)

    def test_json_status(self):
        # Ensure the webui charm is providing us a valid status_urls.json.
        go_to('/static/engine-health/status_urls.json')

        # browser automatically wraps JSON (non-html) into a <pre> ...
        pre = get_element_by_css('pre')
        json.loads(pre.text)

    def test_status_page(self):
        # Show we can load engine-health information.

        # XXX cprov 20141105: status page can be slow (> 30s) if the cache
        # is cold, we should probably work on this more agressively to
        # provide quicker information.

        go_to('/status/')
        assert_title('Ubuntu CI Engine')

        # 'Engine health' is the active link.
        active_link = get_element_by_css('a.active')
        self.assertEqual('Engine health', active_link.text)


if __name__ == "__main__":
    unittest.main()
