"use strict";

// Make jslint happy
var YUI;

YUI.add('webuiforms', function (Y) {

    /**
     * Overriding Y.SelectField for properly rendering '<label>'s.
     */
    Y.WebUISelectField = Y.Base.create(
        'select-field', Y.SelectField, [Y.WidgetParent, Y.WidgetChild], {

        _renderLabelNode : function () {
            var contentBox = this.get('contentBox'),
                titleNode = Y.Node.create('<label></label>');
            titleNode.set('innerHTML', this.get('label'));
            contentBox.appendChild(titleNode);
            this._labelNode = titleNode;
        }
    });

    Y.webuiforms = {

        /**
         * Hardcoded array of series choices, must be kept in sync with
         * the supported options in Ticket-System.
         */
        seriesChoices: [
            {label : 'Vivid - 15.04', value : 'vivid'},
            {label : 'Utopic - 14.10', value : 'utopic'},
            {label : 'Trusty - 14.04', value : 'trusty'},
            {label : 'Precise - 12.04', value : 'precise'}
        ],

        /**
         * Convert form 'sources' and 'merge_proposals' to 'subtickets'.
         *
         * @method _mangleFormSources
         */
        _mangleFormSources: function (form_data) {
            var sources_list = form_data['sources'].split('\n');
            var mp_list = form_data['merge_proposals'].split('\n');
            delete form_data['sources'];
            delete form_data['merge_proposals'];

            form_data['subtickets'] = [];

            var mp_by_source = {}
            mp_list.forEach( function(mp_url) {
                if (!Y.Lang.trim(mp_url)) { return; }
                // https://<lp>/~<user>/<source>/<branch>/+merge/<mp_id>
                var source = mp_url.split('/')[4]
                // Append sources only mentioned in MPs.
                if (sources_list.indexOf(source) == -1) {
                    sources_list.push(source);
                }
                // Group MPs records by source.
                if (!mp_by_source.hasOwnProperty(source)) {
                    mp_by_source[source] = [];
                }
                mp_by_source[source].push({'lp_url': mp_url})
            });

            sources_list.forEach( function(name) {
                if (!Y.Lang.trim(name)) { return; }
                form_data['subtickets'].push({
                    'sourcepackage': {'name': name},
                    'merge_proposals': mp_by_source[name] || null,
                    'assignee': form_data['owner']
                });
            });

            return form_data;
        },

        /**
         * Convert form 'citrain' attributes to 'citrain_overlay'.
         *
         * @method _mangleFormCitrain
         */
        _mangleFormCitrain: function (form_data) {
            var citrain_overlay_fields = [
                'comments', 'test_notes', 'landers', 'sync_request',
                'job_url'
            ]
            var citrain_overlay_obj = {}
            citrain_overlay_fields.forEach( function(name) {
                citrain_overlay_obj[name] = form_data[name];
                delete form_data[name];
            });
            form_data['citrain_overlay'] = citrain_overlay_obj;
            return form_data;
        },

        /**
         * Add appropiate reviews to the form.
         *
         * @method _mangleFormReviews
         */
        _mangleFormReviews: function (form_data) {
            form_data['reviews'] = [
                {'review_type': 'Developer',
                 'workflow_step': 'Silo testing'}
            ];
            if (form_data['qa_review']) {
                form_data['reviews'].push({
                    'review_type': 'QA',
                    'workflow_step': 'Silo testing'
                });
            };
            delete form_data['qa_review'];
            return form_data;
        },

        /**
         * Setup a form for creating a new ticket.
         *
         * Uses Y.Form() to build and render the form dynamically and
         * re-implements its submission to POST JSON data (instead of
         * 'urlencoded'). Opens the just-created ticket on success.
         *
         * @method setupCreateForm
         */
        setupCreateForm: function (container) {
            var form_message = Y.Node.create('<div>');
            Y.one(container).appendChild(form_message);

            // Build and render a form for creating a new Ticket.
            var f = new Y.Form({
                boundingBox: container,
                action : '/api/v1/ticket/',
                method : 'post',
                children : [
                    {name: 'title', required: true, label: 'Title:'},
                    {name: 'description', label: 'Description:',
                     type: 'TextareaField'},
                    {name: 'owner', label: 'Owner:',
                     validator: Y.FormField.VALIDATE_EMAIL_ADDRESS},
                    {name: 'landers', required: true, label: 'Landers:'},
                    {name: 'qa_review', label: 'QA sign-off required?',
                     type: 'CheckboxField'},
                    {name: 'sync_request', label: 'Sync request:'},
                    {name: 'series', required: true, label: 'Series:',
                     type: 'WebUISelectField', useDefaultOption: false,
                     choices: this.seriesChoices},
                    {name: 'sources', label: 'Source Packages:',
                     type: 'TextareaField'},
                    {name: 'merge_proposals', label: 'Merge Proposals:',
                     validator: Y.webuiforms.validateMP,
                     type: 'TextareaField'},
                    {name: 'test_notes', required: true, label: 'Test notes:',
                     type: 'TextareaField'},
                    {name: 'comments', label: 'Comments:',
                     type: 'TextareaField'},
                    {name: 'create-button', type: 'SubmitButton',
                     value: 'Create!'}
                ]
            });

            f.render();
            Y.one('.yui3-submit-button-content').appendChild(
                Y.Node.create('<a>')
                    .set('text', 'Cancel')
                    .set('href', '.')
            );

            // Hijack the default (www-urlencoded) form submission and
            // submit its content as JSON.
            Y.one('input[name="create-button"]').on('click', function(e) {
                e.halt();
                if (!f._runValidation()) {
                    return 1;
                }

                var sources_field = Y.one(
                    'textarea[name="sources"]').get('value');
                var merges_field = Y.one(
                    'textarea[name="merge_proposals"]').get('value');
                if (sources_field === "" && merges_field === "") {
                    f.getField('sources').set('error',
                        'Source Packages or Merge Proposals must be supplied');
                    f.getField('merge_proposals').set('error',
                        'Source Packages or Merge Proposals must be supplied');
                    return 1;
                }

                var form_data = f.toJSON();
                form_data = Y.webuiforms._mangleFormSources(form_data);
                form_data = Y.webuiforms._mangleFormCitrain(form_data);
                form_data = Y.webuiforms._mangleFormReviews(form_data);
                form_data['workflow'] = {'name': 'citrain'};

                Y.io(f.get('action'), {
                    method: f.get('method'),
                    sync: true,
                    data: Y.JSON.stringify(form_data),
                    headers: {
                        'Content-Type': 'application/json;',
                    },
                    on: {
                        success: function(tx, r) {
                            var location = r.getResponseHeader('Location');
                            Y.config.win.location.href = location.replace(
                                '/api/v1', '');
                        },
                        failure: function (tx, r) {
                            Y.log('FAILURE: ' + r.responseText);
                            form_message.addClass('form-error')
                            form_message.set(
                                'text', 'Ticket creation failed!')
                        }
                    }
                });
            });
        },

        /**
         * Validate a LP bug ID (locally only).
         *
         * Positive int32 except ZERO (1 - 2147483648)
         *
         * @method setupEditForm
         */
        validateBugId: function(val, field) {
            var filter = /^[\d]*$/;
            if (filter.test(val) === false || parseInt(val, 10) === 0) {
                field.set('error', 'Please enter a valid Launchpad bug ID.');
                return false;
            }
            return true;
        },

        /**
         * Validate MP entries (locally only).
         *
         * @method validateMP
         */
        validateMP: function(val, field) {
            var s_filter = /https:\/\/code\.launchpad\.net\/~[-a-zA-Z0-9@:%._\+~#=]/;
            var e_filter = /\+merge\/[0-9]+$/;
            var mps = val.split(/[\s,]+/);
            for (var i = 0; i < mps.length; i++){
                if (s_filter.test(mps[i]) === false ||
                        e_filter.test(mps[i]) === false) {
                            // ignore newlines at the end
                            if (mps[i] !== ""){
                                field.set('error', 'The entry '+ mps[i] +
                                        ' is not a valid MP URL.');
                                return false;
                            }
                }
            }
            return true;
        },

        /**
         * Setup a form for editing a ticket.
         *
         * @method setupEditForm
         */
        setupEditForm: function (container) {
            var form_message = Y.Node.create('<div>');
            Y.one(container).appendChild(form_message);

            // Build and render a form for editing Ticket.
            var f = new Y.Form({
                boundingBox: container,
                action : '/api/v1/ticket/' + ticket_id + '/',
                method : 'post',
                children : [
                    {name: 'title', required: true, label: 'Title:'},
                    {name: 'owner', label: 'Owner:',
                     validator: Y.FormField.VALIDATE_EMAIL_ADDRESS},
                    {name: 'landers', required: true, label: 'Landers:'},
                    {name: 'bug_id', label: 'Bug id:',
                     validator: Y.webuiforms.validateBugId,
                     validateInline: true},
                    {name: 'sync_request', label: 'Sync request:'},
                    {name: 'sources', label: 'Source Packages:',
                     type: 'TextareaField'},
                    {name: 'merge_proposals', label: 'Merge Proposals:',
                     validator: Y.webuiforms.validateMP,
                     type: 'TextareaField'},
                    {name: 'description', label: 'Description:',
                     type: 'TextareaField'},
                    {name: 'test_notes', required: true, label: 'Test notes:',
                     type: 'TextareaField'},
                    {name: 'comments', label: 'Comments:',
                     type: 'TextareaField'},
                    {name: 'edit-button', type: 'SubmitButton',
                     value: 'Update!'}
                ]
            });

            // Load ticket information from TS and pre-fill the form.
            Y.io('/api/v1/fullticket/' + ticket_id + '/', {
                sync: true,
                on: {
                    success: function (tx, r) {
                        var ticket = Y.JSON.parse(r.responseText);
                        if (ticket.citrain_overlay) {
                            ticket = Y.merge(ticket, ticket.citrain_overlay);
                        }
                        f.each( function(field) {
                            var name = field.get('name');
                            switch (name) {
                            case 'edit-button':
                                return;
                            case 'bug_id':
                                field.set(
                                    'value', (ticket[name] || '').toString());
                                break;
                            case 'sources':
                                var sources = [];
                                ticket.subticket.forEach( function(s) {
                                    sources.push(s.sourcepackage.name);
                                });
                                field.set('value', sources.join('\n'));
                                break;
                            case 'merge_proposals':
                                var mps = [];
                                ticket.subticket.forEach( function(s) {
                                    s.merge_proposals.forEach( function(mp) {
                                        mps.push(mp.lp_url);
                                    });
                                });
                                field.set('value', mps.join('\n'));
                                break;
                            default:
                                field.set('value', ticket[name] || '');
                            }
                        });
                    },
                    failure: function (tx, r, e) {
                        console.log("failure loading fullticket (" +
                                    ticket_id + "): ", tx, r, e);
                    }
                }
            });

            // Render the form and a 'Cancel' link.
            f.render();
            Y.one('.yui3-submit-button-content').appendChild(
                Y.Node.create('<a>')
                    .set('text', 'Cancel')
                    .set('href', '.')
            );

            Y.one('input[name="edit-button"]').on('click', function(e) {
                e.halt();
                if (!f._runValidation()) {
                    return 1;
                }

                var sources_field = Y.one(
                    'textarea[name="sources"]').get('value');
                var merges_field = Y.one(
                    'textarea[name="merge_proposals"]').get('value');
                if (sources_field === "" && merges_field === "") {
                    f.getField('sources').set('error',
                        'Source Packages or Merge Proposals must be supplied');
                    f.getField('merge_proposals').set('error',
                        'Source Packages or Merge Proposals must be supplied');
                    return 1;
                }

                var form_data = f.toJSON();
                form_data = Y.webuiforms._mangleFormSources(form_data);
                form_data = Y.webuiforms._mangleFormCitrain(form_data);
                /* Tastypie integer field do not cope with empty strings,
                   if cleared, 'bug_id' has to be explicitly null. */
                if (!form_data['bug_id'].trim()) {
                    form_data['bug_id'] = null;
                }

                Y.io(f.get('action'), {
                    // XXX cprov 20141204: we only use and support PUT
                    // because of a phantomjs bug (see TicketResource
                    // 'allowed_methods' for the entire story) it should
                    // be PATCH.
                    method: 'put',
                    sync: true,
                    data: Y.JSON.stringify(form_data),
                    headers: {
                        'Content-Type': 'application/json;',
                    },
                    on: {
                        success: function(tx, r) {
                            Y.config.win.location.href = f.get(
                                'action').replace('/api/v1', '');
                        },
                        failure: function (tx, r) {
                            Y.log('FAILURE: ' + r.responseText);
                            form_message.addClass('form-error')
                            form_message.set(
                                'text', 'Ticket update failed!')
                        }
                    }
                });
            });
        }
    };
}, '0.0.1', {requires: ['io', 'node', 'json-parse', 'json-stringify',
                        'gallery-form']});
